from invana_bot.transformers.mongodb import OTManager, ReadFromMongo, WriteToMongoDB
import requests
from twisted.internet import reactor
import logging

logger = logging.getLogger(__name__)


class RunnerEngineBase(object):
    """

    RunnerEngineBase

    """

    def run(self):
        return self.crawl()

    @staticmethod
    def get_index(transformation_id=None, indexes=None):
        for _index in indexes:
            if _index['transformation_id'] == transformation_id:
                return _index

    def index_data(self, index=None, results_cleaned=None):
        collection_name = index['collection_name']
        unique_key_field = index['unique_key']
        mongo_executor = WriteToMongoDB(
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['CONNECTION_URI'],
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['DATABASE_NAME'],
            collection_name,
            unique_key_field,
            docs=results_cleaned)
        mongo_executor.connect()
        mongo_executor.write()

    def trigger_callback(self, callback_config=None):
        callback_id = callback_config.get("callback_id")
        print("Triggering callback_id: {}".format(callback_id))

        url = callback_config.get("url")
        request_type = callback_config.get("request_type", '').lower()
        if request_type == "get":
            req = requests.get(url, headers=callback_config.get("headers", {}), verify=False)
            response = req.text
            print("Triggered callback successfully and callback responded with message :{}".format(response))
        elif request_type == "post":
            req = requests.post(url, json=callback_config.get("payload", {}),
                                headers=callback_config.get("headers", {}), verify=False)
            response = req.text
            print("Triggered callback successfully and callback responded with message :{}".format(response))

    def callback(self, callback_fn=None):
        all_indexes = self.manifest.get('indexes', [])
        if len(all_indexes) == 0:
            print("There are no callback notifications associated with the indexing jobs. So we are Done here.")
        else:
            print("Initiating, sending the callback notifications after the respective transformations ")
            for index in all_indexes:
                index_id = index.get('index_id')
                callback_config = self.get_callback_for_index(index_id=index_id)
                if callback_config:
                    try:
                        self.trigger_callback(callback_config=callback_config)
                    except Exception as e:
                        print("Failed to send callback[{}] with error: {}".format(callback_config.get("callback_id"),
                                                                                  e))

        if callback_fn is None:
            reactor.stop()
        else:
            callback_fn()

    def get_callback_for_index(self, index_id=None):
        callbacks = self.manifest.get("callbacks", [])
        for callback in callbacks:
            callback_index_id = callback.get('index_id')
            if callback_index_id == index_id:
                return callback
        return

    def transform_and_index(self, callback_fn=None):
        """
        This function will handle both tranform and index

        :param callback:
        :return:
        """
        print("transformer started")

        all_transformation = self.manifest.get('transformations', [])

        for transformation in all_transformation:
            print("transformation", transformation)
            transformation_id = transformation['transformation_id']
            transformation_fn = transformation.get('transformation_fn')

            transformation_index_config = self.get_index(transformation_id=transformation_id,
                                                         indexes=self.manifest['indexes'])
            mongo_executor = ReadFromMongo(
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['CONNECTION_URI'],
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['DATABASE_NAME'],
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['COLLECTION_NAME'],
                query={"context.job_id": self.job_id}
            )
            mongo_executor.connect()
            ops = []
            ot_manager = OTManager(ops).process(mongo_executor)
            results = ot_manager.results
            try:
                results_cleaned__ = transformation_fn(results)
                logger.info("Finished the transformation with transformation_id : {}".format(transformation_id))
            except Exception as e:
                logger.error(
                    "Failed the transformation with transformation_id : {} with error :: {}".format(transformation_id,
                                                                                                    e))
                results_cleaned__ = []
            results_cleaned = []
            for result in results_cleaned__:
                if "_id" in result.keys():
                    del result['_id']
                if "context" in result.keys():
                    result['context']['job_id'] = self.job_id
                else:
                    result['context'] = {'job_id': self.job_id}

                results_cleaned.append(result)
            self.index_data(index=transformation_index_config, results_cleaned=results_cleaned)

            print("Total results_cleaned count of job {} is {}".format(self.job_id, results_cleaned.__len__()))

        print("======================================================")
        print("Successfully crawled + transformed + indexed the data.")
        print("======================================================")
        self.callback(callback_fn=callback_fn)



from scrapy.spiders import CrawlSpider
from scrapy.http import Request

import os


class WebCrawlerBase(CrawlSpider):

    def _build_request(self, rule, link):
        headers = {}
        user_agent_header = os.environ.get("WCP_REQUEST_HEADERS_USER_AGENT")
        if user_agent_header:
            headers = {"User-Agent": user_agent_header}
        r = Request(url=link.url, headers=headers, callback=self._response_downloaded)
        r.meta.update(rule=rule, link_text=link.text)
        return r



class ExtractorBase(object):
    pass

    def __init__(self, response=None, extractor=None, parser_id=None):
        if None in [response, extractor, parser_id]:
            raise Exception("Invalid input to the extractor class, response, extractor and parser_id are mandatory")
        self.response = response
        self.extractor = extractor
        self.parser_id = parser_id



import pymongo
from datetime import datetime


class MongoDBPipeline(object):
    def __init__(self,
                 connection_uri=None,
                 database_name=None,
                 collection_name=None):
        # TODO - implement caching the db_clients specific to collection and database
        if None in [connection_uri, database_name, collection_name]:
            raise Exception("connection_uri, database_name, collection_name should be provided.")

        self.db_client = pymongo.MongoClient(connection_uri)
        self.db = self.db_client[database_name]
        self.collection = self.db[collection_name]

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            connection_uri=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get(
                'CONNECTION_URI', None),
            database_name=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get(
                'DATABASE_NAME', None),
            collection_name=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get(
                'COLLECTION_NAME', None),
        )

    def process_item(self, item, spider):
        if self.collection is None:
            raise Exception("pymongo.MongoClient() it not called in the Pipeline, please make the connection first")
        data = dict(item)
        data['updated'] = datetime.now()
        self.collection.insert(data)
        # print("Item added to Storage", data)
        return item



DEFAULT_SETTINGS_BASE = {
    'COMPRESSION_ENABLED': False,  # this will save the data in normal text form, otherwise to bytes form.
    'HTTPCACHE_ENABLED': True,
    'TELNETCONSOLE_PORT': [6023, 6073],

}

MONGODB_SETTINGS = {
    'ITEM_PIPELINES': {'invana_bot.core.storages.mongodb.MongoDBPipeline': 1},
    'HTTPCACHE_STORAGE': "invana_bot.core.httpcache.mongodb.MongoDBCacheStorage",

    'INVANA_BOT_SETTINGS': {
        'HTTPCACHE_STORAGE_SETTINGS': {
            'CONNECTION_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_cache_db",
            'COLLECTION_NAME': "web_link",
            "EXPIRY_TIME": 3600
        },
        'ITEM_PIPELINES_SETTINGS': {
            'CONNECTION_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_data",
            'COLLECTION_NAME': "website_data"
        }
    }
}

ELASTICSEARCH_SETTINGS = {
    'ITEM_PIPELINES': {'invana_bot.core.storages.elasticsearch.ElasticSearchPipeline': 1},
    'HTTPCACHE_STORAGE': "invana_bot.core.httpcache.elasticsearch.ESCacheStorage",

    'INVANA_BOT_SETTINGS': {
        'HTTPCACHE_STORAGE_SETTINGS': {
            'CONNECTION_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_cache_db",
            'COLLECTION_NAME': "web_link",
            "EXPIRY_TIME": 3600
        },
        'ITEM_PIPELINES_SETTINGS': {
            'CONNECTION_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_data",
            'COLLECTION_NAME': "website_data"
        }
    }
}

DEFAULT_SETTINGS = {}
DEFAULT_SETTINGS.update(DEFAULT_SETTINGS_BASE)
DEFAULT_SETTINGS.update(MONGODB_SETTINGS)



from .base import CTIJobGeneratorBase
from invana_bot.engines.cti import CTIFlowRunnerEngine
from datetime import datetime


class CTIJobGenerator(CTIJobGeneratorBase):
    """
        Invana-bot cti job generator
    """

    def create_job(self, cti_manifest=None, context=None, crawler_cls=None):
        if context is None:
            context = {}
        if 'job_id' not in context.keys():
            context['job_id'] = self.job_id
            context['job_started'] = datetime.now()

        settings_from_manifest = cti_manifest.get("settings", {})
        actual_settings = self.settings
        actual_settings['DOWNLOAD_DELAY'] = settings_from_manifest.get("download_delay", 0)
        cti_runner = CTIFlowRunnerEngine(cti_manifest=cti_manifest,
                                   settings=actual_settings,
                                   job_id=self.job_id,
                                   context=context,
                                   crawler_cls=crawler_cls
                                   )
        job, errors = cti_runner.crawl()
        return {"crawler_job": job, "crawler_job_errors": errors, "runner": cti_runner}



def get_crawler_from_list(crawlers=None, crawler_id=None):
    for crawler in crawlers:
        if crawler.get("crawler_id") == crawler_id:
            return crawler
    raise Exception("crawler_id: {} not found in the crawlers data".format(crawler_id))



from invana_bot.extractors.base import ExtractorBase


class ImageExtractor(ExtractorBase):
    def run(self):
        data = {}
        images_data = []
        images_selector = self.response.css("img")
        for selector in images_selector:
            images_data.append(selector.get("src"))
        data[self.parser_id] = images_data
        return data



# from scrapy.spiders import XMLFeedSpider, Spider
# from invana_bot.utils.url import get_domain
# import feedparser
#
#
# class GenericFeedSpider(XMLFeedSpider):
#     name = "generic_feed_spider"
#     itertag = 'item'
#     iterator = 'xml'  # This is actually unnecessary, since it's the default value
#
#     def get_or_none(self, selector):
#         try:
#             return selector.extract_first()
#         except Exception as e:
#             print(e)
#             return None
#
#     def parse_node(self, response, node):
#         title = self.get_or_none(node.select('title/text()'))
#
#         url = self.get_or_none(node.select('link/text()'))
#         description = self.get_or_none(node.select('description/text()'))
#         pub_date = self.get_or_none(node.select('pubDate/text()'))
#         category = self.get_or_none(node.select('category/text()'))
#         # image = node.select('item/media:content/url')
#
#         item = {}
#         item['title'] = title
#         item['url'] = url
#
#         item['pub_date'] = pub_date
#         item['category'] = category
#         item['description'] = description
#         item['domain'] = get_domain(response.url)
#         return item
#
#
# class RSSSpider(Spider):
#     """
#
#     https://gist.github.com/nyov/9c70e780ea80204559d6da5525228702
#
#     """
#     name = "rss"
#
#     def parse_feed(self, feed):
#         """ Parse RSS/Atom feed using feedparser
#         """
#         data = feedparser.parse(feed)
#         if data.bozo:
#             if (hasattr(data.bozo_exception, 'getLineNumber') and
#                     hasattr(data.bozo_exception, 'getMessage')):
#                 line = data.bozo_exception.getLineNumber()
#                 # segment = feed.split('\n')[line - 1]
#             # could still try to return data. not necessarily completely broken
#             return None
#         return data
#
#     def parse(self, response):
#         # parse downloaded content with feedparser (NOT re-downloading with feedparser)
#         feed = self.parse_feed(response.body)
#         if feed:
#             # grab some feed elements
#             # - https://pythonhosted.org/feedparser/common-rss-elements.html
#             # - https://pythonhosted.org/feedparser/common-atom-elements.html
#
#             # ns = feed.namespaces
#             # feed_title = feed.feed.title
#             # feed_link = feed.feed.link
#             # feed_desc = feed.feed.description
#
#             for entry in feed.entries:
#                 # have content?
#                 content = entry.get('content')
#                 if content:
#                     # content = content[0]
#                     content = content[0]['value']
#
#                 item = {
#                     # global feed data
#                     # 'feed_title': feed_title,
#                     # 'feed_link': feed_link,
#                     # 'feed_description': feed_desc,
#                     #
#                     # item entry data
#                     # 'url': response.url,
#                     'url': entry.link,
#                     'title': entry.title,
#                     'domain': get_domain(response.url),
#                     'description': entry.description,
#                     # 'date': entry.published,
#                     # 'date': entry.published_parsed,
#                     'pub_date': entry.updated_parsed,
#
#                     # optional
#                     'content': content,
#                     'type': entry.get('dc_type'),
#                 }
#
#                 yield item



import sys
import os
import yaml


class SingleCrawlerManifestManager(object):
    """



    """
    required_files = ["crawler_manifest.json", "crawler_manifest.py"]

    def __init__(self, config_path=None):
        print("Setting ETI path as: {}".format(config_path))
        self.config_path = config_path

    def import_files(self):
        # print("self.cti_config_path", self.config_path)
        self.current_crawler = yaml.load(open("{}/crawler_manifest.yml".format(self.config_path)))
        sys.path.append(self.config_path)
        import crawler_transformations
        self.cti_transformations_module = crawler_transformations
        # print("cti_manifest is {}".format(self.current_crawler))
        # print("cti_transformations_module is {}".format(self.cti_transformations_module))

    def validate_cti_path_and_files(self):
        errors = []

        try:
            files_in_path = os.listdir(self.config_path)
        except Exception as e:
            errors.append("No such path exist {}".format(self.config_path))
            files_in_path = []
        if errors == 0:
            for required_file in self.required_files:
                if required_file not in files_in_path:
                    errors.append("{} file not in the path {}".format(required_file, self.config_path))
        return errors

    def import_cti_transformations(self):
        for tranformation in self.current_crawler.get("transformations", []):
            method_to_call = getattr(self.cti_transformations_module, tranformation.get("transformation_fn"))
            tranformation['transformation_fn'] = method_to_call

    def get_manifest(self):
        errors = self.validate_cti_path_and_files()
        if len(errors) > 0:
            return None, errors
        self.import_files()
        self.import_cti_transformations()
        return self.current_crawler, errors



from .base import TraversalBase


class PaginationTraversal(TraversalBase):
    """


    """

    def __init__(self, current_page_count=None, traversal_config=None):
        self.current_page_content = current_page_count
        self.traversal_config = traversal_config



from invana_bot.engines.single import SingleCrawlerRunnerEngine
from .base import CTIJobGeneratorBase
from datetime import datetime


class SingleCrawlJobGenerator(CTIJobGeneratorBase):
    """
    InvanaBot single crawler job generator

    """

    def create_job(self,
                   current_crawler=None,
                   context=None,
                   crawler_cls=None):
        if context is None:
            context = {}
        if 'job_id' not in context.keys():
            context['job_id'] = self.job_id
            context['job_started'] = datetime.now()

        settings_from_manifest = current_crawler.get("settings", {})
        actual_settings = self.settings
        actual_settings['DOWNLOAD_DELAY'] = settings_from_manifest.get("download_delay", 0)
        runner = SingleCrawlerRunnerEngine(settings=actual_settings,
                                     current_crawler=current_crawler,
                                     job_id=self.job_id,
                                     context=context,
                                     crawler_cls=crawler_cls)
        job, errors = runner.crawl()
        return {"crawler_job": job, "crawler_job_errors": errors, "runner": runner}



from invana_bot.utils.crawlers import get_crawler_from_list
from invana_bot.engines.single import SingleCrawlerRunnerEngine
from invana_bot.utils.config import validate_cti_config
from .base import RunnerEngineBase


class CTIFlowRunnerEngine(RunnerEngineBase):
    """

    CTIFlowRunnerEngine


    """

    def __init__(self, cti_manifest=None, settings=None, job_id=None, context=None, crawler_cls=None):
        self.manifest = cti_manifest
        self.settings = settings
        self.crawlers = self.manifest['crawlers']
        self.job_id = job_id
        self.context = context
        self.crawler_cls = crawler_cls

    def crawl(self):
        errors = validate_cti_config(self.manifest)
        if len(errors) == 0:

            initial_crawler = get_crawler_from_list(
                crawler_id=self.manifest['init_crawler']['crawler_id'],
                crawlers=self.crawlers
            )
            initial_crawler['start_urls'] = self.manifest['init_crawler']['start_urls']
            crawler_runner = SingleCrawlerRunnerEngine(
                job_id=self.job_id,
                current_crawler=initial_crawler,
                crawlers=self.crawlers,
                context=self.context,
                crawler_cls=self.crawler_cls,
                settings=self.settings
            )
            cti_job = crawler_runner.run()

            return cti_job, errors
        else:
            return None, errors



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import argparse
from invana_bot.jobs.cti import CTIJobGenerator
from invana_bot.jobs.single import SingleCrawlJobGenerator
from invana_bot.settings.default import DEFAULT_SETTINGS
from invana_bot.manifests.cti import CTIManifestManager
from invana_bot.manifests.single import SingleCrawlerManifestManager


def invana_bot_run():
    parser = argparse.ArgumentParser(description='InvanaBot - A web crawler framework that can'
                                                 ' transform websites into datasets with Crawl, '
                                                 'Transform and Index workflow. ')

    parser.add_argument('--path', type=str, default='./', help='FOO!')
    parser.add_argument('--type', type=str, default='cti', help='options are ct or single', choices=("single", "cti"))

    args = parser.parse_args()
    path = os.path.abspath(args.path)
    crawler_type = args.type
    # print("crawler_type", crawler_type)
    if crawler_type == "cti":
        manifest_manager = CTIManifestManager(
            cti_config_path=path
        )
        cti_manifest, errors = manifest_manager.get_manifest()
        if len(errors) == 0:
            crawler_job_generator = CTIJobGenerator(
                settings=DEFAULT_SETTINGS,
            )
            context = cti_manifest.get("context")
            job = crawler_job_generator.create_job(
                cti_manifest=cti_manifest,
                context=context
            )
            crawler_job_generator.start_job(job=job)
        else:
            print("==============================================================")
            print("ERROR : ETI Job Failing with the errors :: {}".format(
                manifest_manager.cti_config_path,
                errors
            ))
            print("==============================================================")
    elif crawler_type == "single":
        manifest_manager = SingleCrawlerManifestManager(
            config_path=path
        )
        crawler_manifest, errors = manifest_manager.get_manifest()
        if len(errors) == 0:
            crawler_job_generator = SingleCrawlJobGenerator(
                settings=DEFAULT_SETTINGS
            )
            context = crawler_manifest.get("context")
            job = crawler_job_generator.create_job(
                current_crawler=crawler_manifest,
                context=context
            )
            crawler_job_generator.start_job(job=job)
        else:
            print("==============================================================")
            print("ERROR : ETI Job Failing with the errors :: {}".format(
                manifest_manager.config_path,
                errors
            ))
            print("==============================================================")



from invana_bot.extractors.base import ExtractorBase
from invana_bot.utils.selectors import get_selector_element


class AllLinksExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = {}
        for selector in self.extractor.get('data_selectors', []):
            _d = get_selector_element(self.response, selector)
            extracted_data[selector.get('selector_id')] = _d
        data[self.parser_id] = extracted_data
        return data


class SameDomainLinkExtractor(ExtractorBase):
    pass


class ForeignDomainLinkExtractor(ExtractorBase):
    pass


class PaginationLinkExtractor(ExtractorBase):
    def run(self):
        data = {}
        extracted_data = {}
        for selector in self.extractor.get('data_selectors', []):
            _d = get_selector_element(self.response, selector)
            extracted_data[selector.get('selector_id')] = _d
        data[self.parser_id] = extracted_data
        return data


class CustomLinkExtractor(ExtractorBase):
    pass



from invana_bot.utils.exceptions import InvalidCrawlerConfig


def validate_config(config=None):
    required_keys = ['start_url', 'data_selectors']
    for key_ in required_keys:
        if key_ not in config.keys():
            raise InvalidCrawlerConfig(
                "Invalid configuration: Required Key {0} not found in the configuration".format(key_))
    # TODO - validate all the data_selectors data aswell
    return True


class InvanaBotConfigValidator(object):
    CRAWLER_EXAMPLE = {
        "crawler_id": "blogs_list",
        "parsers": [],
        "traversals": []
    }
    PARSER_EXAMPLE = {
        "parser_type": "CustomContentExtractor",
        "parser_id": "example_parser",
        "data_selectors": []
    }
    all_errors = []

    def __init__(self, config=None):
        self.config = config

    def log_error(self, error_text):
        self.all_errors.append(error_text)

    def validate_required_fields(self):
        required_keys = ['cti_id', 'init_crawler', 'crawlers']
        for key_ in required_keys:
            if key_ not in self.config.keys():
                self.log_error(
                    "Invalid configuration: Required Key {0} not found in the configuration".format(key_))
        # TODO - validate all the data_selectors data aswell
        # print("all the required fields exist")
        return True

    # @staticmethod
    # def validate_options_fields():
    #     optional_keys = ['transformations', 'indexes', 'callbacks', 'context']
    #

    def validate_selector(self, selector=None):

        selector_attribute = selector.get("selector_attribute")
        if selector_attribute is None:
            raise InvanaBotConfigValidator("field 'selector_attribute' is mandatory for selector object")

        selector_keys = selector.keys()
        if selector_attribute == "element":
            required_fields = ["selector_id", "selector", "child_selectors", "selector_attribute"]
            for field in required_fields:
                if field not in selector_keys:
                    self.log_error(
                        "{} field not found in 'selector_attribute={}' type selector".format(field, selector_attribute))

        else:
            required_fields = ["selector_id", "selector", "selector_type", "selector_attribute"]

            for field in required_fields:
                if field not in selector_keys:
                    self.log_error(
                        "{} field not found in 'selector_attribute!={}' type selector".format(field,
                                                                                              selector_attribute))

    def validate_crawlers(self, crawlers=None):
        # print(crawlers)
        crawler_required_fields = [
            {
                "parser_id": "crawler_id",
                "field_type": str
            },
            {
                "parser_id": "parsers",
                "field_type": list
            }

        ]

        required_parsers_fields = [
            {
                "parser_id": "parser_type",
                "field_type": str
            }
        ]

        for crawler in crawlers:
            crawler_config_keys = crawler.keys()
            # check if all required keys where there.
            for required_field in crawler_required_fields:
                if required_field['parser_id'] in crawler_config_keys:
                    pass
                else:
                    self.log_error(
                        "required field '{}' in crawlers data not found".format(required_field['parser_id']))

                if type(crawler[required_field['parser_id']]) is not required_field['field_type']:
                    self.log_error(
                        "required field '{}' in crawlers should of '{}' data type".format(required_field['parser_id'],
                                                                                          required_field['field_type']))

            """
            making sure parsers data is correct
            """
            for required_field in required_parsers_fields:
                for parser in crawler['parsers']:
                    parser_config_keys = parser.keys()
                    if required_field['parser_id'] in parser_config_keys:
                        pass
                    else:
                        self.log_error(
                            "required field '{}' in parser data not found".format(required_field['parser_id']))

                    if type(parser[required_field['parser_id']]) is not required_field['field_type']:
                        self.log_error(
                            "required field '{}' in parsers should of '{}' data type".format(
                                required_field['parser_id'],
                                required_field['field_type']))

                    if parser['parser_type'] == "CustomContentExtractor":
                        if len(parser.get('data_selectors', [])) == 0:
                            self.log_error(
                                "data_selectors should be specified when using parser_type 'CustomContentExtractor'")

                    data_selectors = parser.get("data_selectors", [])
                    for selector in data_selectors:
                        pass  # TODO - implement data selector validations.
                        self.validate_selector(selector=selector)

            traversals = crawler.get('traversals', [])
            if type(traversals) is not list:
                self.log_error(
                    "traversals data in the crawler '{}' should be of list type".format(crawler['crawler_id']))

            self.validate_traversals(traversals=traversals, crawler=crawler, all_crawlers=crawlers)

    def validate_traversals(self, traversals=None, crawler=None, all_crawlers=None):
        valid_traversal_types = ['pagination', 'same_domain', 'link_from_field']

        all_crawlers_ids = [crawler['crawler_id'] for crawler in all_crawlers]

        for traversal in traversals:
            traversal_type = traversal.get("traversal_type")
            if traversal_type not in valid_traversal_types:
                self.log_error("Traversal of types '{}' are accepted, where as you have "
                               " parser_id '{}'".format(", ".join(valid_traversal_types),
                                                          traversal_type))

            if traversal_type not in traversal.keys():
                self.log_error("Traversal of type '{}' should have '{}' key defining the"
                               " traversal configuration ".format(traversal_type, traversal_type))
            next_crawler_id = traversal.get("next_crawler_id")

            if next_crawler_id is None:
                self.log_error("All Traversals should have next_crawler_id set ")

            if next_crawler_id not in all_crawlers_ids:
                self.log_error("You are using next_crawler_id '{}' "
                               "but it is not defined in the crawlers."
                               " Available crawler_id in the config are {}".format(next_crawler_id,
                                                                                   all_crawlers_ids))

            if traversal_type == "pagination":
                required_fields = ['selector', 'selector_type']
                traversal_example = {
                    "traversal_type": "pagination",
                    "pagination": {
                        "sel ector": ".next-posts-link",
                        "selector_type": "css",
                        "max_pages": 4
                    },
                    "next_crawler_id": "blogs_list"
                }
                for required_field in required_fields:
                    if required_field not in traversal[traversal_type].keys():
                        self.log_error(
                            "Traversal type '{}' in crawler '{}' should have the config with keys '{}' along with optional max_pages."
                            " Example: {}"
                            "".format(
                                traversal_type,
                                crawler['crawler_id'],
                                ", ".join(required_fields), traversal_example))

            elif traversal_type == "same_domain":
                pass  # don't have any required fields.

            elif traversal_type == "link_from_field":
                traversal_example = {
                    "traversal_type": "link_from_field",
                    "link_from_field": {
                        "parser_id": "blog_list_parser",
                        "selector_id": "url"
                    },
                    "next_crawler_id": "blog-detail"
                }
                required_fields = ['parser_id', 'selector_id']
                for required_field in required_fields:
                    if required_field not in traversal[traversal_type].keys():
                        self.log_error(
                            "Traversal type '{}' in crawler '{}' should have the config "
                            "with keys '{}' along with optional max_pages. Example: {}"
                            "".format(
                                traversal_type,
                                crawler['crawler_id'],
                                ", ".join(required_fields), traversal_example))

    def validate_transformations_and_indexes(self):
        transformations = self.config.get('transformations', [])
        indexes = self.config.get('indexes', [])
        if len(indexes) == 0:
            if len(transformations) > 0:
                self.log_error("transformations cannot be applied if indexes is not defined.")
            print("Ignoring the transformation if index")
        else:
            transformation_ids = [transformation.get('transformation_id') for transformation in transformations]

            for transformation in transformations:
                transformation_id = transformation.get("transformation_id")
                transformation_fn = transformation.get("transformation_fn")
                if transformation_fn is None:
                    self.log_error("transformation with transformation_id "
                                   "'{}' doesn't have transformation_fn set".format(transformation_id))

            example_index = {
                "transformation_id": "default",
                "connection_uri": "mongodb://127.0.0.1/crawlers_data_index",
                "collection_name": "blogs_list",
                "unique_key": "url"
            }
            # transformation_ids_in_indexes = [index.get('transformation_id') for index in indexes]
            index_required_fields = example_index.keys()
            for index in indexes:
                transformation_id = index.get("transformation_id")
                for required_field in index_required_fields:
                    if index.get(required_field) is None:
                        self.log_error("required field '{}' is missing in the index configuration."
                                       " Example is {} ".format(required_field, example_index))

                if transformation_id not in transformation_ids:
                    self.log_error("index with transformation_id '{}' has no"
                                   " transformation defined in transformations".format(transformation_id))

    def validate_callback(self):
        all_index_ids = [index.get('index_id') for index in self.config.get("indexes", [])]

        callback_template = {
            "callback_id": "default",
            "index_id": "default",
            "url": "http://localhost/api/callback",
            "request_type": "POST",
            "payload": {
            },
            "headers": {
                "X-TOKEN": "abc123456789"
            }
        }

        callbacks = self.config.get("callbacks", [])
        callback_required_keys = callback_template.keys()
        for callback in callbacks:
            callback_id = callback.get("callback_id")
            callback_index_id = callback.get("index_id")
            if callback_id is None:
                self.log_error("Invalid callback configuration, callback_id cannot be None."
                               " Here is the example {}".format(callback_template))

            if callback_index_id not in all_index_ids:
                self.log_error("callback is set with an index - '{}' which is not "
                               "defined in the indexes configuration."
                               "".format(callback_index_id))

            for required_field in callback_required_keys:
                if callback.get(required_field) is None:
                    self.log_error("Required key {} missing in callback_id '{}' "
                                   "".format(required_field, callback_id))

    def validate_settings(self):
        settings = self.config.get("settings", {})
        required_keys = ['download_delay', 'allowed_domains']
        for required_key in required_keys:
            if required_key not in settings.keys():
                self.log_error(
                    "required field '{}' in settings data not found".format(required_key))

    def validate(self):
        self.validate_required_fields()
        self.validate_crawlers(crawlers=self.config['crawlers'])
        self.validate_transformations_and_indexes()
        self.validate_callback()
        self.validate_settings()
        return self.all_errors


# def validate_cti_config(config=None):
#     optional_keys = ['transformations', 'indexes', 'callbacks']
#     required_keys = ['cti_id', 'init_crawler', 'crawlers']
#     for key_ in required_keys:
#         if key_ not in config.keys():
#             raise InvalidCrawlerConfig(
#                 "Invalid configuration: Required Key {0} not found in the configuration".format(key_))
#     # TODO - validate all the data_selectors data aswell
#     return True


# def process_config(config=None):
#     processed_config_dict = {}
#     parent_selectors = []
#     new_config_selectors = []
#     for selector in config.get('data_selectors', []):
#         if selector.get('selector_attribute') == 'element':
#             parent_selectors.append(selector)
#
#     """
#     process the element root selectors (without parent_selector) which might become dictionaries
#     """
#     if len(parent_selectors) > 0:
#         for parent_selector in parent_selectors:
#             processed_config_dict[parent_selector.get('id')] = parent_selector
#             processed_config_dict[parent_selector.get('id')]['child_selectors'] = []
#
#             for selector in config['data_selectors']:
#                 if selector.get('parent_selector') == parent_selector.get('id'):
#                     processed_config_dict[parent_selector.get('id')]['child_selectors'].append(selector)
#
#     """
#     process the elements with no root selectors
#     """
#     for selector in config.get('data_selectors', []):
#         if selector.get('parent_selector') is None and selector.get('selector_attribute') != 'element':
#             new_config_selectors.append(selector)
#
#     for k, v in processed_config_dict.items():
#         new_config_selectors.append(v)
#     config['data_selectors'] = new_config_selectors
#     return config


def validate_cti_config(cti_manifest):
    validator = InvanaBotConfigValidator(config=cti_manifest)
    errors = validator.validate()
    return errors


def validate_crawler_config(crawler_config):
    # TODO - fix this validations for crawler config later.
    return []



import uuid
from scrapy.utils.log import configure_logging
from twisted.internet import reactor
from scrapy import signals
from scrapy.crawler import Crawler, CrawlerRunner
from scrapy.settings import Settings


class CTIJobGeneratorBase(object):
    """


    """
    settings = {
        'COMPRESSION_ENABLED': False,  # this will save the data in normal text form, otherwise to bytes form.
        'HTTPCACHE_ENABLED': True,
        'TELNETCONSOLE_PORT': None

    }
    runner = None

    def __init__(self,
                 job_id=None,
                 settings=None,
                 **kwargs):

        if settings is None:
            raise Exception("settings should be set")
        self.settings = settings
        if job_id is None:
            self.job_id = self.generate_job_id()
        else:
            self.job_id = job_id
        self.set_logger()

    @staticmethod
    def set_logger():
        configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})

    @staticmethod
    def generate_job_id():
        return uuid.uuid4().hex

    def _validate_urls(self, urls):
        if type(urls) is None:
            raise Exception("urls should be list type.")
        if len(urls) is 0:
            raise Exception("urls length should be atleast one.")

    def get_settings(self):
        return self.settings

    def start_job(self, job=None, callback_fn=None):
        runner = CrawlerRunner()
        crawler_job = job['crawler_job']
        cti_runner = job['runner']
        crawler_cls = crawler_job['crawler_cls']
        crawler_kwargs = crawler_job['crawler_kwargs']

        def engine_stopped_callback():
            cti_runner.transform_and_index(callback_fn=callback_fn)

        if callback_fn:
            print("""
==========================================================
WARNING: callback_fn is {}
==========================================================
Since start_job is called with callback_fn, make sure you end the reactor if you want the crawler process to
stop after the callback function is executed. By default callback_fn=None will close the reactor.

To write a custom callback_fn

def callback_fn():
    print ("Write your own callback logic")
    from twisted.internet import reactor
    reactor.stop()
==========================================================
        """.format(callback_fn))

        crawler = Crawler(crawler_cls, Settings(cti_runner.settings))
        crawler.signals.connect(engine_stopped_callback, signals.engine_stopped)
        runner.crawl(crawler, **crawler_kwargs)
        """
        d = runner.crawl(crawler, **crawler_kwargs)
        # d.addBoth(engine_stopped_callback)
        """
        reactor.run()


"""
class InvanaFeedCrawler(CTIJobGeneratorBase):

    def crawl_feeds(self, feed_urls=None):
        self.setup_crawler_type_settings(crawler_type="feeds")
        self._validate_urls(feed_urls)

        process = CrawlerProcess(self.settings)
        allowed_domains = []
        for feed_url in feed_urls:
            domain = feed_url.split("://")[1].split("/")[0]  # TODO - clean this
            allowed_domains.append(domain)

        process.crawl(RSSSpider,
                      start_urls=feed_urls,
                      )
        process.start()

"""



from invana_bot.extractors.base import ExtractorBase
from invana_bot.utils.selectors import get_selector_element


class ParagraphsExtractor(ExtractorBase):
    def run(self):
        data = {}
        paragraphs_data = []
        elements = self.response.css("p").extract()
        for el in elements:
            paragraphs_data.append(el)
        data[self.parser_id] = {}
        data[self.parser_id]['paragraphs'] = paragraphs_data
        return data


class HeadingsExtractor(ExtractorBase):
    # TODO - implement this
    pass


class MainContentExtractor(ExtractorBase):
    # TODO - implement this
    pass


class TableContentExtractor(ExtractorBase):
    def run(self):
        data = {}
        tables = []
        for table in self.response.css("table"):
            table_data = []
            table_headers = [th.extract() for th in table.css("thead tr th::text")]
            for row in table.css("tbody tr"):
                row_data = [td.extract() for td in row.css("td::text")]
                row_dict = dict(zip(table_headers, row_data))
                table_data.append(row_dict)
            tables.append(table_data)
        data[self.parser_id] = {}
        data[self.parser_id]['tables'] = tables
        return data


class HTMLMetaTagExtractor(ExtractorBase):
    def run(self):
        data = {}
        meta_data_dict = {}
        elements = self.response.css('meta')
        for element in elements:
            meta_property = element.xpath("@{0}".format('property')).extract_first()
            if meta_property:
                meta_property = meta_property.replace(":", "__")
                meta_data_dict[meta_property] = element.xpath("@{0}".format('content')).extract_first()

        data[self.parser_id] = meta_data_dict
        return data


class CustomContentExtractor(ExtractorBase):

    def run(self):
        data = {}
        extracted_data = {}
        for selector in self.extractor.get('data_selectors', []):
            if selector.get('selector_attribute') == 'element' and len(selector.get('child_selectors', [])) > 0:
                # TODO - currently only support multiple elements strategy. what if multiple=False
                elements = self.response.css(selector.get('selector'))
                elements_data = []
                for el in elements:
                    datum = {}
                    for child_selector in selector.get('child_selectors', []):
                        _d = get_selector_element(el, child_selector)
                        datum[child_selector.get('selector_id')] = _d if _d else None
                    elements_data.append(datum)
                data_type = selector.get("data_type", "RawField")
                if data_type.startswith("List") is False:
                    single_data = elements_data[0]
                    extracted_data[selector.get('selector_id')] = single_data
                else:
                    extracted_data[selector.get('selector_id')] = elements_data
            else:
                _d = get_selector_element(self.response, selector)
                extracted_data[selector.get('selector_id')] = _d
        data[self.parser_id] = extracted_data
        return data



from datetime import datetime
from elasticsearch_dsl import DocType, Date, Integer, Text, connections
from invana_bot.utils.url import get_urn


class ElasticSearchPipeline(object):
    """


    pipeline_settings = {
        'ITEM_PIPELINES': {'invana_bot.storages.elasticsearch.ElasticSearchPipeline': 1},
        'HTTPCACHE_STORAGE': "invana_bot.httpcache.elasticsearch.ESCacheStorage",
    }

    es_settings = {
        'INVANA_BOT_SETTINGS': {
            'HTTPCACHE_STORAGE_SETTINGS': {
                'CONNECTION_URI': "127.0.0.1",
                'DATABASE_NAME': "crawler_cache_db",
                'COLLECTION_NAME': "web_link",
                "EXPIRY_TIME": 3600
            },
            'ITEM_PIPELINES_SETTINGS': {
                'CONNECTION_URI': "127.0.0.1",
                'DATABASE_NAME': "crawler_data",
                'COLLECTION_NAME': "crawler_feeds_data"
            }
        }
    }

    """

    def setup_collection(self):
        class WebLinkExtracted(DocType):
            url = Text()
            body = Text()
            headers = Text()
            status = Integer()
            created = Date()

            class Meta:
                index = self.database_name
                doc_type = self.collection_name

        return WebLinkExtracted

    def __init__(self, connection_uri=None,
                 database_name=None,
                 collection_name=None):
        self.connection_uri = connection_uri
        self.database_name = database_name
        self.collection_name = collection_name
        connections.create_connection(hosts=[self.connection_uri])
        self.WebLinkExtracted = self.setup_collection()
        self.WebLinkExtracted.init()

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            connection_uri=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get('CONNECTION_URI'),
            database_name=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get(
                'DATABASE_NAME'),
            collection_name=crawler.settings.get('INVANA_BOT_SETTINGS').get('ITEM_PIPELINES_SETTINGS').get(
                'COLLECTION_NAME')
        )

    def _flatten_headers(self, obj):  # TODO -may be not using !!
        flat_data = {}
        for k, v in obj.items():
            flat_data['headers_{}'.format(k)] = v
        return flat_data

    def process_item(self, item, spider):
        data = dict(item)
        data['updated'] = datetime.now()
        self.WebLinkExtracted(meta={'id': get_urn(data['url'])}, **data).save()
        return item



from invana_bot.crawlers.default import InvanaBotSingleWebCrawler
from scrapy.spiders import Rule
from invana_bot.utils.config import validate_crawler_config
from scrapy.linkextractors import LinkExtractor
from .base import RunnerEngineBase


class SingleCrawlerRunnerEngine(RunnerEngineBase):
    """

SingleCrawlerRunnerEngine
    crawler = {  # single pipe

        "crawler_id": "blog-list",
        "parsers": [
            {
                "data_selectors": [
                    {
                        "id": "items",
                        "selector": ".post-listing .post-item",
                        "selector_attribute": "element",
                        "multiple": True
                    }
                ],
            }
        ],
        "traversals": [{
            "traversal_type": "pagination",
            "pagination": {
                "selector": ".next-posts-link",
                "selector_type": "css",
                "max_pages": 2
            },
        }],
        "context":{
        },
        "settings": {
        }

    }

    """

    def __init__(self,
                 current_crawler=None,
                 crawlers = None,
                 job_id=None,
                 context=None,
                 crawler_cls=None,
                 settings=None
                 ):
        """

        :param current_crawler: single crawler in the CTI flow
        :param crawlers: all the crawlers in the CTI flow
        :param context: any extra information user want to send to the crawled data or carry forward.
        :param settings:  settings to run the crawling job. .
        """
        self.manifest = current_crawler
        self.job_id = job_id
        if crawlers is None:
            self.crawlers = [current_crawler]
        else:
            self.crawlers = crawlers
        self.settings = settings
        self.crawler_cls = crawler_cls
        if context:
            self.context = context

    def crawl(self):
        errors = validate_crawler_config(self.manifest)
        if len(errors) == 0:

            cti_job = self.run()

            return cti_job, errors
        else:
            return None, errors

    def validate_pipe(self):
        must_have_keys = ["crawler_id", "parsers"]
        optional_keys = ["traversals"]
        for key in must_have_keys:
            if key not in self.manifest.keys():
                raise Exception(
                    "invalid parser data, should have the following keys; {}".format(",".join(must_have_keys)))

    def validate_traversal(self):
        pass  # TODO - implement this

    def validate_extractor(self):
        pass  # TODO - implement this

    def get_traversals(self):
        return self.manifest.get("traversals", [])

    def get_extractors(self):
        return self.manifest.get("parsers", [])

    def generate_crawler_kwargs(self):
        extractor = LinkExtractor()
        rules = [
            Rule(extractor, follow=True)  # TODO - add regex types of needed.
        ]
        allowed_domains = self.manifest.get("settings", {}).get('allowed_domains', [])
        crawler_kwargs = {
            "start_urls": self.manifest['start_urls'],
            "allowed_domains": allowed_domains,
            "rules": rules,
            "current_crawler": self.manifest,
            "crawlers": self.crawlers,
            "context": self.context
        }
        return crawler_kwargs

    def run(self):
        crawler_cls = self.crawler_cls or InvanaBotSingleWebCrawler

        crawler_kwargs = self.generate_crawler_kwargs()
        return {"crawler_cls": crawler_cls, "crawler_kwargs": crawler_kwargs, "spider_settings": self.settings}



from .base import WebCrawlerBase
from invana_bot.extractors.content import CustomContentExtractor, \
    ParagraphsExtractor, TableContentExtractor, HTMLMetaTagExtractor
from invana_bot.extractors.links import PaginationLinkExtractor
import scrapy
from invana_bot.utils.url import get_domain, get_absolute_url
from invana_bot.utils.crawlers import get_crawler_from_list
from urllib.parse import urlparse

TRAVERSAL_LINK_FROM_FIELD = "link_from_field"
TRAVERSAL_SAME_DOMAIN_FIELD = "same_domain"


class InvanaBotSingleWebCrawler(WebCrawlerBase):
    """
    This is generic spider
    """
    name = "InvanaBotSingleWebCrawler"

    def closed(self, reason):
        print("spider closed with payload:", reason, self.current_crawler)

    @staticmethod
    def run_extractor(response=None, extractor=None):
        parser_type = extractor.get("parser_type")
        parser_id = extractor.get("parser_id")
        if parser_type in [None, "CustomContentExtractor"]:
            extractor_object = CustomContentExtractor(response=response, extractor=extractor,
                                                      parser_id=parser_id)
        elif parser_type == "TableContentExtractor":
            extractor_object = TableContentExtractor(response=response, extractor=extractor,
                                                     parser_id=parser_id or "tables")

        elif parser_type == "PaginationLinkExtractor":
            extractor_object = PaginationLinkExtractor(response=response, extractor=extractor,
                                                       parser_id=parser_id or "pagination")

        elif parser_type == "HTMLMetaTagExtractor":
            extractor_object = HTMLMetaTagExtractor(response=response, extractor=extractor,
                                                    parser_id=parser_id or "meta_tags")
        elif parser_type == "ParagraphsExtractor":
            extractor_object = ParagraphsExtractor(response=response, extractor=extractor,
                                                   parser_id=parser_id or "paragraphs")
        else:
            return {}
        data = extractor_object.run()
        return data

    @staticmethod
    def get_subdocument_key(crawler=None, parser_id=None):
        """
        element is the subdocument key name.

        :param crawler:
        :param parser_id:
        :param selector_id:
        :return:
        """
        for extractor in crawler['parsers']:
            if extractor.get("parser_id") == parser_id:
                for selector in extractor.get('data_selectors', []):
                    if selector.get('selector_attribute') == 'element':
                        return selector.get("selector_id")
        return

    def post_parse(self, response=None):
        pass

    def parse(self, response=None):

        current_crawler = response.meta.get("current_crawler")
        crawlers = response.meta.get("crawlers")
        context = self.context

        if None in [crawlers, current_crawler]:
            current_crawler = self.current_crawler
            crawlers = self.crawlers

        data = {}
        for extractor in current_crawler['parsers']:
            extracted_data = self.run_extractor(response=response, extractor=extractor)
            data.update(extracted_data)

        if context is not None:
            data.update({"context": context})
        data['url'] = response.url
        data['domain'] = get_domain(response.url)
        data['context']['crawler_id'] = current_crawler['crawler_id']
        yield data

        for traversal in current_crawler.get('traversals', []):
            if traversal['traversal_type'] == "pagination":
                # TODO - move this to run_pagination_traversal(self, response=None, traversal=None) method;
                traversal_config = traversal['pagination']
                next_crawler_id = traversal['next_crawler_id']
                max_pages = traversal_config.get("max_pages", 1)
                current_page_count = response.meta.get('current_page_count', 1)
                if current_page_count < max_pages:
                    next_selector = traversal_config.get('selector')
                    if next_selector:
                        if traversal_config.get('selector_type') == 'css':
                            next_page = response.css(next_selector + "::attr(href)").extract_first()
                        elif traversal_config.get('selector_type') == 'xpath':
                            next_page = response.xpath(next_selector + "::attr(href)").extract_first()
                        else:
                            next_page = None
                        current_page_count = current_page_count + 1
                        if next_page:
                            if not "://" in next_page:
                                next_page_url = "https://" + get_domain(response.url) + next_page
                            else:
                                next_page_url = next_page
                            next_crawler = get_crawler_from_list(crawler_id=next_crawler_id, crawlers=crawlers)
                            yield scrapy.Request(
                                next_page_url,
                                callback=self.parse,
                                meta={
                                    "current_page_count": current_page_count,
                                    "current_crawler": next_crawler,
                                    "crawlers": crawlers,
                                }
                            )
            elif traversal['traversal_type'] == TRAVERSAL_LINK_FROM_FIELD:
                next_crawler_id = traversal['next_crawler_id']
                traversal_config = traversal[TRAVERSAL_LINK_FROM_FIELD]

                subdocument_key = self.get_subdocument_key(
                    crawler=current_crawler,
                    parser_id=traversal_config['parser_id']
                )

                for item in data.get(traversal_config['parser_id']).get(subdocument_key, []):
                    traversal_url = item[traversal[TRAVERSAL_LINK_FROM_FIELD]['selector_id']]
                    if traversal_url:
                        if "://" not in traversal_url:  # TODO - fix this monkey patch
                            url_parsed = urlparse(response.url)
                            traversal_url = url_parsed.scheme + "://" + url_parsed.netloc + "/" + traversal_url.lstrip(
                                "/")

                        next_crawler = get_crawler_from_list(crawler_id=next_crawler_id, crawlers=crawlers)
                        yield scrapy.Request(
                            traversal_url,
                            callback=self.parse,
                            meta={
                                "crawlers": crawlers,
                                "current_crawler": next_crawler,
                            }
                        )
                    else:
                        print("ignoring traversal to {}".format(traversal_url))
            elif traversal['traversal_type'] == TRAVERSAL_SAME_DOMAIN_FIELD:
                all_urls = response.css("a::attr(href)").extract()
                filtered_urls = []
                all_urls = list(set(all_urls))
                current_domain = get_domain(response.url)
                for url in all_urls:
                    url = get_absolute_url(url=url, origin_url=response.url)
                    if get_domain(url) == current_domain:
                        filtered_urls.append(url)
                filtered_urls = list(set(filtered_urls))
                # max_pages = traversal.get("max_pages", 100)
                #  implementing max_pages is difficult cos it keeps adding
                # new 100 pages in each thread.
                current_page_count = response.meta.get('current_page_count', 1)
                next_crawler_id = traversal['next_crawler_id']
                next_parser = get_crawler_from_list(crawler_id=next_crawler_id, crawlers=crawlers)

                for url in filtered_urls:
                    current_page_count = current_page_count + 1

                    yield scrapy.Request(
                        url, callback=self.parse,
                        meta={
                            "current_page_count": current_page_count,
                            "current_crawler": next_parser,
                            "crawlers": crawlers
                        }
                    )
        self.post_parse(response=response)
